@@ -22,6 +22,7 @@ global_config = get_driver().config
 
 from .openai_func import TextGenerator
 from .Extension import Extension
+from .text_func import cos_sim
 
 global_data = {}  # 用于存储所有数据的字典
 global_data_path = f"{config['NG_DATA_PATH']}naturel_gpt.pkl"
@@ -30,7 +31,7 @@ global_extensions = {}  # 用于存储所有扩展的字典
 
 
 """ ======== 定义会话类 ======== """
-# 对话类
+# 会话类
 class Chat:
     preset_key = ''  # 预设标识
     chat_presets = None         # 当前对话预设
@@ -42,7 +43,7 @@ class Chat:
 
     def __init__(self, chat_key:str, preset_key:str = ''):
         self.chat_presets = {}   # 当前对话预设
-        self.chat_preset_dicts = {}   # 预设字典
+        self.chat_preset_dicts = {} # 预设字典
         self.chat_key = chat_key    # 对话标识
         self.is_enable = True       # 启用会话
         if not preset_key:  # 如果没有预设，选择默认预设
@@ -110,6 +111,40 @@ class Chat:
             logger.info(f"印象生成消耗token数: {tg.cal_token_count(prompt + global_preset_userdata[self.preset_key][userid]['chat_impression'])}")
             global_preset_userdata[self.preset_key][userid]['chat_history'] = global_preset_userdata[self.preset_key][userid]['chat_history'][-config['CHAT_MEMORY_SHORT_LENGTH']:]
 
+    # 为当前预设设置记忆
+    def set_memory(self, mem_key:str, mem_value:str = '') -> None:
+        if 'chat_memory' not in self.chat_presets:
+            self.chat_presets['chat_memory'] = {}
+
+        # 如果没有指定mem_value，则删除该记忆
+        if not mem_value:
+            if mem_key in self.chat_presets['chat_memory']:
+                del self.chat_presets['chat_memory'][mem_key]
+                if config.get('__DEBUG__'): logger.info(f"忘记了: {mem_key}")
+            else:
+                logger.warning(f"尝试删除不存在的记忆 {mem_key}")
+        else:   # 否则设置该记忆，并将其移到在最后
+            if mem_key in self.chat_presets['chat_memory']:
+                del self.chat_presets['chat_memory'][mem_key]
+            self.chat_presets['chat_memory'][mem_key] = mem_value
+            if config.get('__DEBUG__'): logger.info(f"记住了: {mem_key} -> {mem_value}")
+
+            if len(self.chat_presets['chat_memory']) > config['CHAT_MEMORY_MAX_LENGTH']:   # 检查记忆是否超过最大长度 超出则删除最早的记忆并记录日志
+                del_key = list(self.chat_presets['chat_memory'].keys())[0]
+                del self.chat_presets['chat_memory'][del_key]
+                if config.get('__DEBUG__'): logger.info(f"忘记了: {del_key} (超出最大记忆长度)")
+
+    # 增强记忆 如果响应中的内容与记忆中的内容相似，则增强记忆(将改记忆移到最后)
+    def enhance_memory(self, response:str) -> bool:
+        if 'chat_memory' not in self.chat_presets:
+            return
+        for mem_key, mem_value in self.chat_presets['chat_memory'].items():#模糊匹配
+            if cos_sim(response, mem_value) > config['CHAT_MEMORY_ENHANCE_THRESHOLD']:
+                self.set_memory(mem_key, mem_value)
+                if config.get('__DEBUG__'): logger.info(f"增强记忆 {mem_key}")
+                return True
+        return False
+
     # 修改对话预设
     def change_presettings(self, preset_key:str) -> None:
         if preset_key not in self.chat_preset_dicts:    # 如果聊天预设字典中没有该预设，则从全局预设字典中拷贝一个
@@ -140,6 +175,42 @@ class Chat:
         impression_text = f"[impression]\n{global_preset_userdata[self.preset_key][userid].get('chat_impression')}\n\n" \
             if global_preset_userdata[self.preset_key].get(userid, {}).get('chat_impression', None) else ''  # 用户印象描述
 
+        # 记忆模块
+        memory_text = ''
+        memory = ''
+        if 'chat_memory' not in self.chat_presets:
+            self.chat_presets['chat_memory'] = {}
+        if self.chat_presets.get('chat_memory', {}):  # 删除空记忆
+            self.chat_presets['chat_memory'] = {k: v for k, v in self.chat_presets['chat_memory'].items() if v}
+        if self.chat_presets.get('chat_memory', {}):    # 如果有记忆，则生成记忆模板
+            idx = 0 # 记忆序号
+            for k, v in self.chat_presets['chat_memory'].items():
+                idx += 1
+                memory_text += f"{idx}. {k}: {v}\n"
+
+        # 删除多余的记忆
+        if len(self.chat_presets.get('chat_memory', {})) > config.get('MEMORY_MAX_LENGTH', 16):
+            self.chat_presets['chat_memory'] = {k: v for k, v in sorted(self.chat_presets['chat_memory'].items(), key=lambda item: item[1])}
+            self.chat_presets['chat_memory'] = {k: v for k, v in list(self.chat_presets['chat_memory'].items())[:config.get('MEMORY_MAX_LENGTH', 16)]}
+            logger.info(f"删除多余记忆: {self.chat_presets['chat_memory']}")
+
+        if config.get('MEMORY_ACTIVE'):  # 如果记忆功能开启
+            if global_extensions.get('remember') and global_extensions.get('forget'): # 如果记忆功能已加载
+                memory = (  # 如果有记忆，则生成记忆模板
+                    f"[history memory (max length: {config.get('MEMORY_MAX_LENGTH', 16)} - Please delete the unimportant memory in time before exceed it)]\n"
+                    f"{memory_text}\n"
+                    f"ATTENTION: The earlier chat history may not be provided again in the next request, use /#remember&key&value#/ to remember something\n\n"
+                ) if memory_text else ( # 如果没有记忆，则生成空记忆模板
+                    f"[memory (max length: {config.get('MEMORY_MAX_LENGTH', 16)} - Delete the unimportant memory in time before exceed it)]\n"
+                    f"ATTENTION: The earlier chat history may not be provided again in the next request. There are currently no saved memories, use /#remember&key&value#/ to remember something.\n\n"
+                )
+            else:   # 如果没有加载 memory 拓展，则使用固定记忆
+                logger.warning("未加载主动记忆 memory 拓展，仅启用固定记忆！")
+                memory = (
+                    f"[history memory]\n"
+                    f"{memory_text}\n"
+                ) if memory_text else ''
+
         # 对话历史
         offset = 0
         chat_history:str = '\n\n'.join(self.chat_presets['chat_history'][-(config['CHAT_MEMORY_SHORT_LENGTH'] + offset):])  # 从对话历史中截取短期对话
@@ -157,44 +228,74 @@ class Chat:
         ext_descs = ''.join([global_extensions[ek].generate_description(chat_history) for ek in global_extensions.keys()])
         # 拓展使用示例
         extension_text = (
-            '[Extension options]\n'
+            f"[Extension functions: You can use the following extension functions. The extension module can be invoked multiple times in a single response.]\n"
             # 'Including the above content in a chat message will call the extension module for processing.\n'
             # 'importrant: The extension option is available if and only if in the following strict format. Multiple options can be used in one response.\n'
             # '- Random > min:a; max:b (send a random number between a and b)'
-            'Following the following format in the response will invoke the extension module for the corresponding implementation. The extension module can be invoked multiple times in a single response.\n'
+            # 'Following the following format in the response will invoke the extension module for the corresponding implementation. The extension module can be invoked multiple times in a single response.\n'
             f'{ext_descs}\n'
-            'Usage format in response: /#extension_name&param1&param2#/ (parameters are separated by &)\n'
-            'ATTENTION: Do not use any extensions in response that are not listed above!\n'
+            "Usage format in response: /#{extension_name}&{param1}&{param2}#/ (parameters are separated by '&')\n"
+            'ATTENTION: Do not use any extensions in response that are not listed above! If the response contains content in this format, the extension will be called directly for execution. Do not respond any content in this format if you do not want to call the extension\n'
             # 'example use in response: i will send 2 random number /#Random&0&5#/ /#Random&5&10#/\n\n'    # 拓展使用示例 /#拓展名&参数1&参数2#/，参数之间用&分隔
         ) if config.get('NG_ENABLE_EXT') and ext_descs else (
-            '[Extension options]\n'
-            'No extension is currently available. Do not use the extension function like /#extension_name&param1&param2#/.\n'
+            '[Extension response options]\n'
+            'No extension is currently available. Do not use the extension function like /#{extension_name}&{param1}&{param2}#/.\n'
         )
 
         # 发言提示
-        say_prompt = f"(Multiple segment replies are separated by '*;', single quotes are not included, please only give the details of {self.chat_presets['bot_name']} response and do not give any irrelevant information)" if config.get('NG_ENABLE_MSG_SPLIT') else ''
+        # say_prompt = f"(Multiple segment replies are separated by '*;', single quotes are not included, please only give the details of {self.chat_presets['bot_name']} response and do not give any irrelevant information)" if config.get('NG_ENABLE_MSG_SPLIT') else ''
 
         res_rule_prompt = (
-            f"\n[Response rule: Please follow the following rules strictly]\n"
-            f"\n1. If you need to generate multiple replies, use '*;' delimited(single quotes are not included)"
+            f"\n[Response rule: Your response needs to follow the following rules]\n"
+            f"\n1. If the content of a reply is too long, please segment it in the appropriate place, use '*;' delimited(single quotes are not included)"
             # f"\n2. Only give the response content of {self.chat_presets['bot_name']} and do not carry any irrelevant information or the speeches of other members"
-            f"\n2. Please play the {self.chat_presets['bot_name']} role and only give the reply content of the {self.chat_presets['bot_name']} role, response needs to follow the role's setting and habits"
-            f"\n3. If the response contains code blocks, use the markdown format below"
+            # f"\n2. Please play the {self.chat_presets['bot_name']} role and only give the reply content of the {self.chat_presets['bot_name']} role, response needs to follow the role's setting and habits(Provided by the user)"
+            f"\n2. If the response contains code blocks, use the markdown format below"
             f"\n```python"
             f"\nprint('hi')"
             f"\n```"
         )
 
+        # # 返回对话 prompt 模板
+        # return (    # 返回对话 prompt 模板
+        #     f"[Character setting]"
+        #     f"\n{self.chat_presets['bot_self_introl']}"
+        #     f"\n{summary}\n{impression_text}\n{memory}"
+        #     f"{extension_text}"
+        #     f"{res_rule_prompt}"
+        #     f"\n[Chat History (current time: {time.strftime('%Y-%m-%d %H:%M:%S')})]\n"
+        #     f"\n{chat_history}\n{self.chat_presets['bot_name']}:"
+        # )
+
         # 返回对话 prompt 模板
-        return (    # 返回对话 prompt 模板
-            f"[Character setting]"
-            f"\n{self.chat_presets['bot_self_introl']}"
-            f"\n{summary}\n{impression_text}"
-            f"{extension_text}"
-            f"{res_rule_prompt}"
-            f"\n[Chat - current time: {time.strftime('%Y-%m-%d %H:%M:%S')}]\n"
-            f"\n{chat_history}\n{self.chat_presets['bot_name']}:"
-        )
+        return [
+            {'role': 'system', 'content': ( # 系统消息
+                # f"You must strictly follow the user's instructions to give {self.chat_presets['bot_name']}'s response."
+                f"You must play the specified role according to the instructions of the user and give the response information in accordance with the changed role. When necessary, you can generate a reply in a specified format to invoke extension functions"
+                f"\n{extension_text}"
+                f"\n{res_rule_prompt}"
+            )},
+            {'role': 'user', 'content': (   # 用户消息(演示场景)
+                f"[Character setting]\nAI is an assistant robot.\n\n"
+                # "[memory (max length: 16 - Delete the unimportant memory in time before exceed it)]"
+                f"[history memory (max length: {config.get('MEMORY_MAX_LENGTH', 16)} - Please delete the unimportant memory in time before exceed it)]\n"
+                "\n1. Developer's email: developer@mail.com\n"
+                "\n[Chat History (current time: 2023-03-05 16:29:45)]\n"
+                "\nDeveloper: my email is developer@mail.com, remember it!\n"
+                "\nAI: ok, I will remember it /#remember&Developer's email&developer@mail.com#/\n"
+                "\nDeveloper: Send an email to me for testing\n"
+                "\nAI:(waiting for response...)"
+            )},
+            {'role': 'assistant', 'content': (  # 助手消息(演示输出)
+                "ok, I will send an email, please wait a moment /#email&example@mail.com&test title&hello this is a test#/ *; I have sent an e-mail. Did you get it?"
+            )},
+            {'role': 'user', 'content': (   # 用户消息(实际场景)
+                f"[Character setting]\n{self.chat_presets['bot_self_introl']}\n\n"
+                f"{memory}{impression_text}{summary}"
+                f"\n[Chat History (current time: {time.strftime('%Y-%m-%d %H:%M:%S')})]\n"
+                f"\n{chat_history}\n{self.chat_presets['bot_name']}:(waiting for response...)"
+            )},
+        ]
 
     # 获取当前对话bot的名称
     def get_chat_bot_name(self) -> str:
@@ -644,6 +745,62 @@ async def _(event: Event, arg: Message = CommandArg()):
             chat_info += f"+ {chat.generate_description()}"
         await identity.finish((f"当前已加载的会话:\n{chat_info}"))
 
+    elif cmd in ["记忆", "memory"] and len(cmd.split(' ')) == 1:
+        # 检查主动记忆拓展模块和主动记忆功能是否启用
+        if not (global_extensions.get('remember') and global_extensions.get('forget') and config.get('MEMORY_ACTIVE')):
+            logger.warning("记忆拓展模块未启用或主动记忆功能未开启！")
+        # 检查权限
+        if str(event.user_id) not in config['ADMIN_USERID']:
+            await identity.finish("对不起！你没有权限进行此操作 ＞﹏＜")
+        # 生成记忆信息
+        memory_info = ''
+        for k, v in chat_dict[chat_key].chat_presets.get('chat_memory', {}).items():
+            memory_info += f"+ {k}: {v}\n"
+        if memory_info == '':
+            memory_info = "当前的记忆是空的呢 >﹏<"
+        await identity.finish((f"当前人格记忆:\n{memory_info}"))
+
+    elif cmd in ["记忆", "memory"] and len(cmd.split(' ')) == 2:
+        # 检查主动记忆拓展模块和主动记忆功能是否启用
+        if not (global_extensions.get('remember') and global_extensions.get('forget') and config.get('MEMORY_ACTIVE')):
+            logger.warning("记忆拓展模块未启用或主动记忆功能未开启！")
+        # 检查权限
+        if str(event.user_id) not in config['ADMIN_USERID']:
+            await identity.finish("对不起！你没有权限进行此操作 ＞﹏＜")
+
+    elif cmd.split(' ')[0] in ["记忆", "memory"] and len(cmd.split(' ')) == 3:
+        # 检查主动记忆拓展模块和主动记忆功能是否启用
+        if not (global_extensions.get('remember') and global_extensions.get('forget') and config.get('MEMORY_ACTIVE')):
+            logger.warning("记忆拓展模块未启用或主动记忆功能未开启！")
+        # 检查权限
+        if str(event.user_id) not in config['ADMIN_USERID']:
+            await identity.finish("对不起！你没有权限进行此操作 ＞﹏＜")
+        # 检查是否存在记忆
+        if 'chat_memory' not in chat_dict[chat_key].chat_presets:
+            chat_dict[chat_key].chat_presets['chat_memory'] = {}
+        # 检查记忆是否存在
+        if cmd.split(' ')[1] not in chat_dict[chat_key].chat_presets['chat_memory']:
+            await identity.finish("当前人格没有此记忆！")
+        # 检查操作
+        if cmd.split(' ')[2] in ["删除", "del", "delete"]:
+            chat_dict[chat_key].set_memory(cmd.split(' ')[1], None)
+            await identity.finish(f"已删除记忆: {cmd.split(' ')[1]}")
+
+    elif cmd.split(' ')[0] in ["记忆", "memory"] and len(cmd.split(' ')) == 4:
+        # 检查主动记忆拓展模块和主动记忆功能是否启用
+        if not (global_extensions.get('remember') and global_extensions.get('forget') and config.get('MEMORY_ACTIVE')):
+            logger.warning("记忆拓展模块未启用或主动记忆功能未开启！")
+        # 检查权限
+        if str(event.user_id) not in config['ADMIN_USERID']:
+            await identity.finish("对不起！你没有权限进行此操作 ＞﹏＜")
+        # 检查是否存在记忆
+        if 'chat_memory' not in chat_dict[chat_key].chat_presets:
+            chat_dict[chat_key].chat_presets['chat_memory'] = {}
+        # 检查操作
+        if cmd.split(' ')[1] in ["编辑", "edit"]:
+            chat_dict[chat_key].set_memory(cmd.split(' ')[2], cmd.split(' ')[3])
+            await identity.finish(f"编辑记忆: {cmd.split(' ')[2]} 成功! (￣▽￣)")
+
     else:
         await identity.finish("输入的命令好像有点问题呢... 请检查下再试试吧！ ╮(>_<)╭")
 
@@ -748,7 +905,9 @@ async def do_msg_response(trigger_userid:str, trigger_text:str, is_tome:bool, ma
 
     # 生成对话 prompt 模板
     prompt_template = chat.get_chat_prompt_template(userid=trigger_userid)
-    if config.get('__DEBUG__'): logger.info("对话 prompt 模板: \n" + prompt_template)
+    # 生成 log 输出用的 prompt 模板
+    log_prompt_template = '\n'.join([f"[{m['role']}]\n{m['content']}\n" for m in prompt_template]) if isinstance(prompt_template, list) else prompt_template
+    if config.get('__DEBUG__'): logger.info("对话 prompt 模板: \n" + str(log_prompt_template))
 
     raw_res, success = await tg.get_response(prompt=prompt_template, type='chat', custom={'bot_name': chat.get_chat_bot_name(), 'sender_name': sender_name})  # 生成对话结果
     if not success:  # 如果生成对话结果失败，则直接返回
@@ -766,9 +925,9 @@ async def do_msg_response(trigger_userid:str, trigger_text:str, is_tome:bool, ma
     # 提取后去除所有markdown格式的代码块，剩余部分为对话结果
     talk_res = re.sub(r"```(.+?)```", '', raw_res)
 
-    # 分割对话结果提取出所有 "/#拓展名&参数1&参数2#/" 格式的拓展调用指令 参数之间用&分隔
-    ext_calls = re.findall(r"/.?#(.+?)#.?/", talk_res)
-    # 提取后去除所有拓展调用指令，剩余部分为对话结果
+    # 分割对话结果提取出所有 "/#拓展名&参数1&参数2#/" 格式的拓展调用指令 参数之间用&分隔 多行匹配
+    ext_calls = re.findall(r"/.?#(.+?)#.?/", talk_res, re.S)
+    # 提取后去除所有拓展调用指令，剩余部分为对话结果 多行匹配
     talk_res = re.sub(r"/.?#(.+?)#.?/", '', talk_res)
 
     # 对分割后的对话根据 '*;' 进行分割，表示对话结果中的分句，处理结果为列表，其中每个元素为一句话
@@ -832,7 +991,7 @@ async def do_msg_response(trigger_userid:str, trigger_text:str, is_tome:bool, ma
         else:
             for key in reply:   # 遍历回复内容类型字典
                 if key == 'text' and reply.get(key) and reply.get(key).strip(): # 发送文本
-                    await matcher.send(reply.get(key))
+                    await matcher.send(reply.get(key).strip())
                 elif key == 'image' and reply.get(key): # 发送图片
                     await matcher.send(MessageSegment.image(file=reply.get(key, '')))
                     logger.info(f"回复图片消息: {reply.get(key)}")
@@ -841,13 +1000,21 @@ async def do_msg_response(trigger_userid:str, trigger_text:str, is_tome:bool, ma
                     await matcher.send(Message(MessageSegment.record(file=reply.get(key), cache=0)))
                 elif key == 'code_block' and reply.get(key):  # 发送代码块
                     await matcher.send(Message(reply.get(key)))
+                elif key == 'memory' and reply.get(key):  # 记忆存储
+                    logger.info(f"存储记忆: {reply.get(key)}")
+                    chat.set_memory(reply.get(key).get('key'), reply.get(key).get('value'))
+                    if config.get('__DEBUG__'):
+                        if reply.get(key).get('key') and reply.get(key).get('value'):
+                            await matcher.send(f"[debug]: 记住了 {reply.get(key).get('key')} = {reply.get(key).get('value')}")
+                        elif reply.get(key).get('key') and reply.get(key).get('value') is None:
+                            await matcher.send(f"[debug]: 忘记了 {reply.get(key).get('key')}")
 
                 res_times -= 1
                 if res_times < 1:  # 如果回复次数超过限制，则跳出循环
                     break
         await asyncio.sleep(1.5)  # 每条回复之间间隔1.5秒
 
-    cost_token = tg.cal_token_count(prompt_template + raw_res)  # 计算对话结果的 token 数量
+    cost_token = tg.cal_token_count(str(prompt_template) + raw_res)  # 计算对话结果的 token 数量
 
     while time.time() - sta_time < 1.5:   # 限制对话响应时间
         time.sleep(0.1)
